//! Fayalite Simulation

use crate::{
    bundle::{Bundle, BundleType},
    enum_::Enum,
    expr::Expr,
    int::Bool,
    intern::{Intern, Interned},
    module::{
        Block, Module, ModuleBody, NormalModuleBody, Stmt, StmtConnect, StmtFormal, StmtIf,
        StmtMatch,
    },
    sim::interpreter::{Insn, Insns, InsnsBuilding},
    source_location::SourceLocation,
};
use hashbrown::HashMap;
use std::{cell::RefCell, rc::Rc};

mod interpreter;

#[derive(Debug)]
struct CompilingModule {
    compiled_module: Option<CompiledModule>,
}

#[derive(Debug, PartialEq, Eq, Hash, Clone, Copy)]
enum CondStack {
    Always,
    IfTrue {
        parent: Interned<CondStack>,
        cond: Expr<Bool>,
        source_location: SourceLocation,
    },
    IfFalse {
        parent: Interned<CondStack>,
        cond: Expr<Bool>,
        source_location: SourceLocation,
    },
    MatchArm {
        parent: Interned<CondStack>,
        enum_expr: Expr<Enum>,
        variant_index: usize,
        source_location: SourceLocation,
    },
}

#[derive(Debug)]
pub struct Compiler {
    insns: Insns<InsnsBuilding>,
    modules: HashMap<Interned<Module<Bundle>>, Rc<RefCell<CompilingModule>>>,
    module_queue: Vec<Interned<Module<Bundle>>>,
}

impl Compiler {
    pub fn new() -> Self {
        Self {
            insns: Insns::new(),
            modules: HashMap::new(),
            module_queue: Vec::new(),
        }
    }
    pub fn add_module(&mut self, module: Interned<Module<Bundle>>) {
        self.modules.entry(module).or_insert_with(|| {
            self.module_queue.push(module);
            Rc::new(RefCell::new(CompilingModule {
                compiled_module: None,
            }))
        });
    }
    fn compile_block(
        &mut self,
        module: Interned<Module<Bundle>>,
        block: Block,
        cond_stack: Interned<CondStack>,
    ) {
        let Block { memories, stmts } = block;
        for memory in memories {
            todo!("implement memory");
        }
        for stmt in stmts {
            match stmt {
                Stmt::Connect(StmtConnect {
                    lhs,
                    rhs,
                    source_location,
                }) => todo!(),
                Stmt::Formal(StmtFormal {
                    kind,
                    clk,
                    pred,
                    en,
                    text,
                    source_location,
                }) => todo!("implement simulating formal statements"),
                Stmt::If(StmtIf {
                    cond,
                    source_location,
                    blocks: [then_block, else_block],
                }) => {
                    self.compile_block(
                        module,
                        then_block,
                        CondStack::IfTrue {
                            parent: cond_stack,
                            cond,
                            source_location,
                        }
                        .intern_sized(),
                    );
                    self.compile_block(
                        module,
                        else_block,
                        CondStack::IfFalse {
                            parent: cond_stack,
                            cond,
                            source_location,
                        }
                        .intern_sized(),
                    );
                }
                Stmt::Match(StmtMatch {
                    expr,
                    source_location,
                    blocks,
                }) => {
                    for (variant_index, block) in blocks.into_iter().enumerate() {
                        self.compile_block(
                            module,
                            block,
                            CondStack::MatchArm {
                                parent: cond_stack,
                                enum_expr: expr,
                                variant_index,
                                source_location,
                            }
                            .intern_sized(),
                        );
                    }
                }
                Stmt::Declaration(stmt_declaration) => todo!(),
            }
        }
    }
    fn compile_module(&mut self, module: Interned<Module<Bundle>>) -> CompiledModule {
        let step_entry_pc = self.insns.next_insn_pc();
        match module.body() {
            ModuleBody::Normal(NormalModuleBody { body }) => {
                self.compile_block(module, body, CondStack::Always.intern_sized())
            }
            ModuleBody::Extern(_extern_module_body) => {
                todo!("simulating extern module: {}", module.name_id());
            }
        }
        self.insns.push(Insn::Return, module.source_location());
        CompiledModule { step_entry_pc }
    }
    pub fn run(mut self) -> Compiled {
        while let Some(module) = self.module_queue.pop() {
            let compiling_module = self.modules[&module].clone();
            let mut compiling_module = compiling_module.borrow_mut();
            let CompilingModule { compiled_module } = &mut *compiling_module;
            assert!(compiled_module.is_none());
            *compiled_module = Some(self.compile_module(module));
        }
        Compiled {
            insns: Insns::from(self.insns).intern_sized(),
            modules: self
                .modules
                .into_iter()
                .map(|(module, compiling_module)| {
                    (
                        module,
                        Rc::into_inner(compiling_module)
                            .expect("only reference is Compiler::modules")
                            .into_inner()
                            .compiled_module
                            .expect("module is compiled by Compiler::run"),
                    )
                })
                .collect(),
        }
    }
}

#[derive(Debug)]
struct CompiledModule {
    step_entry_pc: usize,
}

#[derive(Debug)]
pub struct Compiled {
    insns: Interned<Insns>,
    modules: HashMap<Interned<Module<Bundle>>, CompiledModule>,
}

impl Compiled {
    pub fn new<T: BundleType>(module: Module<T>) -> Self {
        let mut compiler = Compiler::new();
        compiler.add_module(module.canonical().intern());
        compiler.run()
    }
}

pub struct Simulation {
    state: interpreter::State,
}
